
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <Mowitz/MwUtils.h>
#include "MwHtmlParser.h"
#include <Mowitz/MwFormat.h>

enum {LIST_NONE=0, LIST_UL, LIST_OL, LIST_DL};

typedef struct hstate {
	int pre;			/* preformatted, default 0 = false */
	int list;			/* list mode, default LIST_NONE */
	int ll;				/* list level, starts with 1 */
	int indent;			/* indentation level, default 0 */
	int fmt;			/* format */
	int (*put)(void *, int, int);	/* function to print one character */
	void (*begin_a)(void *, char *);
	void (*end_a)(void *);
	void (*img)(void *, char *);
	void *closure;
} hstate;

static int href;	/* links, default 0 */
static char **hrefs;

static MwFmt fmt0 = {
        "Times",    /* font family */
        120,            /* size in decipoints */
        0,              /* bold */
        0,              /* italic */
        0,              /* underline */
        0,              /* strikethrough */
        "black",        /* foreground */
        "white",        /* background */
        0,              /* borders */
        MW_VADJ_CENTER, /* vertical adjust */
        MW_HADJ_LEFT,   /* horizontal adjust */
        0};             /* style */

static MwFmt fmt_h1 = {"Helvetica", 240, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h2 = {"Helvetica", 180, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h3 = {"Helvetica", 160, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h4 = {"Helvetica", 140, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h5 = {"Helvetica", 120, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h6 = {"Helvetica", 100, 1, 0, 0, 0, "black", "white", 0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};

static void tag_a(MwHtmlNode *, hstate *);
static void tag_address(MwHtmlNode *, hstate *);
static void tag_area(MwHtmlNode *, hstate *);
static void tag_b(MwHtmlNode *, hstate *);
static void tag_body(MwHtmlNode *, hstate *);
static void tag_br(MwHtmlNode *, hstate *);
static void tag_dd(MwHtmlNode *, hstate *);
static void tag_dl(MwHtmlNode *, hstate *);
static void tag_dt(MwHtmlNode *, hstate *);
static void tag_h1(MwHtmlNode *, hstate *);
static void tag_h2(MwHtmlNode *, hstate *);
static void tag_h3(MwHtmlNode *, hstate *);
static void tag_h4(MwHtmlNode *, hstate *);
static void tag_h5(MwHtmlNode *, hstate *);
static void tag_h6(MwHtmlNode *, hstate *);
static void tag_head(MwHtmlNode *, hstate *);
static void tag_hr(MwHtmlNode *, hstate *);
static void tag_html(MwHtmlNode *, hstate *);
static void tag_i(MwHtmlNode *, hstate *);
static void tag_img(MwHtmlNode *, hstate *);
static void tag_li(MwHtmlNode *, hstate *);
static void tag_map(MwHtmlNode *, hstate *);
static void tag_ol(MwHtmlNode *, hstate *);
static void tag_p(MwHtmlNode *, hstate *);
static void tag_pre(MwHtmlNode *, hstate *);
static void tag_table(MwHtmlNode *, hstate *);
static void tag_td(MwHtmlNode *, hstate *);
static void tag_th(MwHtmlNode *, hstate *);
static void tag_tr(MwHtmlNode *, hstate *);
static void tag_ul(MwHtmlNode *, hstate *);
static void tag_unknown(MwHtmlNode *, hstate *);
static void text(MwHtmlNode *, hstate *);

static struct {
        char *name;
        void (*action)(MwHtmlNode *, hstate *);
} tag[] = {
	{"A", tag_a},
	{"ADDRESS", tag_address},
	{"AREA", tag_area},
	{"B", tag_b},
	{"BR", tag_br},
	{"BODY", tag_body},
	{"DD", tag_dd},
	{"DL", tag_dl},
	{"DT", tag_dt},
	{"H1", tag_h1},
	{"H2", tag_h2},
	{"H3", tag_h3},
	{"H4", tag_h4},
	{"H5", tag_h5},
	{"H6", tag_h6},
	{"HEAD", tag_head},
	{"HR", tag_hr},
	{"HTML", tag_html},
	{"I", tag_i},
	{"IMG", tag_img},
	{"LI", tag_li},
	{"MAP", tag_map},
	{"OL", tag_ol},
	{"P", tag_p},
	{"PRE", tag_pre},
	{"TABLE", tag_table},
	{"TD", tag_td},
	{"TH", tag_th},
	{"TR", tag_tr},
	{"UL", tag_ul},
        {NULL, tag_unknown}
};

static void pprintf(void *cl, int (*put)(void *, int, int),
	int f, const char *fmt, ...)
{
	va_list ap;
	char b[1024];
	int i;

	va_start(ap, fmt);
	vsprintf(b, fmt, ap);
	for (i = 0; b[i]; i++) put(cl, f, b[i]);
	va_end(ap);
}

/* p looks like this: key=value key2="value 2". Keys are not case
   sensitive. Everything not in quotes is converted to upper case.
   Returns value if found, otherwise NULL.
*/
char *MwHtmlGetValue(char *k, char *p)
{
	char *q = p;
	int n = strlen(k);
	int i;
	char b[1024];

Again:
	while (*q && isspace(*q)) q++;	/* skip leading space */
	if (!strncasecmp(k, q, n) && q[n] == '=') {	/* gotcha */
		n++;
		i = 0;
		if (q[n] == '"') {
			n++;
			while (q[n] && q[n] != '"') b[i++] = q[n++];
		} else {
			while (q[n] && !isspace(q[n])) b[i++] = q[n++];
		}
		b[i] = '\0';
		return MwStrdup(b);
	}
	while (*q && !isspace(*q)) q++;
	if (*q) goto Again;
	return NULL;
}

static int store_href(hstate *s, char *p)
{
	char *u = MwHtmlGetValue("HREF", p);
	if (u) {
		href++;
		hrefs = MwRealloc(hrefs, (href+1) * sizeof *hrefs);
		hrefs[href] = u;
		return href;
	}
	return 0;
}

static hstate *new_hstate(int fmt)
{
	hstate *s = MwMalloc(sizeof *s);
	s->pre = 0;
	s->indent = 0;
	s->list = LIST_NONE;
	s->ll = 0;
	s->fmt = fmt;
	s->put = NULL;
	s->closure = NULL;
	return s;
}

static void free_hstate(hstate *s)
{
	MwFree(s);
}

static void newline(hstate *s)
{
	int i;

	s->put(s->closure, s->fmt, '\n');
	for (i = 0; i < s->indent; i++) s->put(s->closure, s->fmt, ' ');
}

static void tree(MwHtmlNode *h, hstate *s)
{
	void (*action)(MwHtmlNode *, hstate *);
	int i;

	if (h == NULL) return;
	if (h->tag == NULL) {
		text(h, s);
	} else {
		for (i = 0; tag[i].name; i++) {
			if (!strcmp(tag[i].name, h->tag)) break;
		}
		action = tag[i].action;
		(*action)(h, s);
	}
}

/* tag handlers are in alphabetical order so I can find them... */

/* Anchor tag. */
static void tag_a(MwHtmlNode *h, hstate *s)
{
	MwFmt fmt;
	hstate ns = *s;
	MwDecodeFormat(ns.fmt, ~0, &fmt);
	fmt.uline = 1;
	fmt.fg = "blue";
	ns.fmt = MwEncodeFormat(~0, &fmt);
	if (store_href(s, h->text) != 0)
#if 0
		pprintf(s->closure, s->put, s->fmt, "[%d]", href);
#else
;
#endif
	if (s->begin_a) (*s->begin_a)(s->closure, h->text);
	tree(h->child, &ns);
	if (s->end_a) (*s->end_a)(s->closure);
	tree(h->next, s);
}

/* Address tag. Only changes font. */
static void tag_address(MwHtmlNode *h, hstate *s)
{
	MwFmt fmt;
	hstate ns = *s;
	MwDecodeFormat(ns.fmt, ~0, &fmt);
	fmt.italic = 1;
	ns.fmt = MwEncodeFormat(~0, &fmt);
	tree(h->child, &ns);
	tree(h->next, s);
}

/* Treat these as <a> */
static void tag_area(MwHtmlNode *h, hstate *s)
{
	if (store_href(s, h->text) != 0)
		pprintf(s->closure, s->put, s->fmt, "[%d]", href);
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_b(MwHtmlNode *h, hstate *s)
{
	MwFmt fmt;
	hstate ns = *s;
	MwDecodeFormat(ns.fmt, ~0, &fmt);
	fmt.bold = 1;
	ns.fmt = MwEncodeFormat(~0, &fmt);
	tree(h->child, &ns);
	tree(h->next, s);
}

/* Check for background values et al */
static void tag_body(MwHtmlNode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Line break */
static void tag_br(MwHtmlNode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* description in description list */
static void tag_dd(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	tree(h->next, &ns);
}

/* start description list */
static void tag_dl(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.list = LIST_DL;
	ns.ll = 1;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* tag in description list */
static void tag_dt(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.indent -= 4;
	newline(&ns);
	tree(h->child, &ns);
	tree(h->next, &ns);
}

/* all headers set font (unimplemented), start a new line, print
   their children and start a new line
*/
static void tag_h1(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h1);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

static void tag_h2(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h2);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

static void tag_h3(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h3);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

static void tag_h4(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h4);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

static void tag_h5(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h5);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

static void tag_h6(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	newline(s);
	ns.fmt = MwEncodeFormat(~0, &fmt_h6);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

/* ignore this tag and all children */
static void tag_head(MwHtmlNode *h, hstate *s)
{
	tree(h->next, s);
}

/* Horizontal ruler */
static void tag_hr(MwHtmlNode *h, hstate *s)
{
	newline(s);
	pprintf(s->closure, s->put, s->fmt,
		"------------------------------------------------------------");
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* this tag doesn't do anything per se */
static void tag_html(MwHtmlNode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_i(MwHtmlNode *h, hstate *s)
{
	MwFmt fmt;
	hstate ns = *s;
	MwDecodeFormat(ns.fmt, ~0, &fmt);
	fmt.italic = 1;
	ns.fmt = MwEncodeFormat(~0, &fmt);
	tree(h->child, &ns);
	tree(h->next, s);
}

/* Image tag. Ignored until we have somewhere to show pictures. */
static void tag_img(MwHtmlNode *h, hstate *s)
{
	char b[1024];
	char *p = MwHtmlGetValue("ALT", h->text);
	if (p == NULL) strcpy(b, "[INLINE]");
	else strcpy(b, p);
	MwFree(p);
	pprintf(s->closure, s->put, s->fmt, "%s", b);
	tree(h->child, s);
	tree(h->next, s);
}

/* start a new line. Should also fiddle with indentation */
static void tag_li(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.indent -= 2;
	newline(&ns);
	switch (ns.list) {
	case LIST_UL:
		pprintf(s->closure, s->put, s->fmt, "* ");
		break;
	case LIST_OL:
		pprintf(s->closure, s->put, s->fmt, "%d. ", ns.ll);
		break;
	default:
		pprintf(s->closure, s->put, s->fmt, "  ");
	}
	ns.indent += 2;
	ns.ll++;
	tree(h->child, &ns);
	tree(h->next, &ns);
}

/* Not much to do here... */
static void tag_map(MwHtmlNode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_ol(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.list = LIST_OL;
	ns.ll = 1;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* Paragraph break */
static void tag_p(MwHtmlNode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* Set a flag that this is preformatted. */
static void tag_pre(MwHtmlNode *h, hstate *s)
{
	MwFmt fmt;
	hstate ns = *s;
	MwDecodeFormat(ns.fmt, ~0, &fmt);
	fmt.family = "Courier";
	ns.fmt = MwEncodeFormat(~0, &fmt);
	ns.pre = 1;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* We don't really do tables just yet. */
static void tag_table(MwHtmlNode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Table cell. One complication is that cells on the same row can be
   siblings or children in the parse tree, and that is correct HTML.
*/
static void tag_td(MwHtmlNode *h, hstate *s)
{
	s->put(s->closure, s->fmt, '\t');
	tree(h->child, s);
	tree(h->next, s);
}

/* Table header */
static void tag_th(MwHtmlNode *h, hstate *s)
{
	s->put(s->closure, s->fmt, '\t');
	tree(h->child, s);
	tree(h->next, s);
}

/* Table row */
static void tag_tr(MwHtmlNode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* To do anything useful with this, we need some state to tell us
   we're in a list and what kind the list is. IOW, add a struct
   hstate or something. Also add indentation level for children.
*/
static void tag_ul(MwHtmlNode *h, hstate *s)
{
	hstate ns = *s;
	ns.list = LIST_UL;
	ns.ll = 1;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* Handles tags we don't have handlers for. Ignore the tag, do children */
static void tag_unknown(MwHtmlNode *h, hstate *s)
{
	/* ignore the tag */
	tree(h->child, s);
	tree(h->next, s);
}

/* Copy text onto screen. Treats whitespace differently if this
   is "preformatted".
*/
static void text(MwHtmlNode *h, hstate *s)
{
	int i;
	int space_pending = 0;
	if (s->pre) {
		for (i = 0; h->text[i]; i++) {
			s->put(s->closure, s->fmt, h->text[i]);
		}
	} else {
		for (i = 0; h->text[i]; i++) {
			if (isspace(h->text[i])) space_pending = 1;
			else {
				if (space_pending) {
					s->put(s->closure, s->fmt, ' ');
					space_pending = 0;
				}
				s->put(s->closure, s->fmt, h->text[i]);
			}
		}
	}
#if 0
#if 1
	for (i = 0; h->text[i]; i++) {
		if (s->pre) s->put(s->closure, s->fmt, h->text[i]);
		else if (isspace(h->text[i])) s->put(s->closure, s->fmt, ' ');
		else s->put(s->closure, s->fmt, h->text[i]);
	}
#else
	for (i = 0; h->text[i]; i++) {
		s->put(s->closure, s->fmt, h->text[i]);
	}
#endif
#endif
	/* can't have children */
	tree(h->next, s);
}

void MwHtmlDumpTree(MwHtmlNode *h, void *closure,
	int (*put)(void *, int, int),
	void (*begin_a)(void *, char *),
	void (*end_a)(void *),
	void (*img)(void *, char *))
{
	int i;

	int f = MwEncodeFormat(~0, &fmt0);
	hstate *s = new_hstate(f);
	s->closure = closure;
	s->put = put;
	s->begin_a = begin_a;
	s->end_a = end_a;
	s->img = img;
	if (hrefs) {	/* free old junk */
		for (i = 1; i <= href; i++)
			MwFree(hrefs[i]);
		MwFree(hrefs);
		href = 0;
		hrefs = NULL;
	}
	tree(h, s);
	newline(s);
#if 0
	if (href) {
		pprintf(s->closure, s->put, s->fmt, "References:\n\n");
		for (i = 1; i <= href; i++)
			pprintf(s->closure, s->put, s->fmt, "[%d] %s\n", i, hrefs[i]);
	}
#endif
	free_hstate(s);
}

int MwHtmlGetHref(char *url, int n)
{
	if (n < 1 || n > href) return 0;
	strcpy(url, hrefs[n]);
	return 1;
}

